package cz.jzitnik.chronos.controllers;

import com.fasterxml.jackson.databind.ObjectMapper;
import cz.jzitnik.chronos.entities.Item;
import cz.jzitnik.chronos.entities.Message;
import cz.jzitnik.chronos.entities.MessageType;
import cz.jzitnik.chronos.entities.Player;
import cz.jzitnik.chronos.payload.errors.ItemNotUsableException;
import cz.jzitnik.chronos.payload.errors.NotFoundError;
import cz.jzitnik.chronos.payload.requests.PlayerNameRequest;
import cz.jzitnik.chronos.payload.responses.InventoryFullResponse;
import cz.jzitnik.chronos.payload.responses.UnifiedResponse;
import cz.jzitnik.chronos.repository.GameRepository;
import cz.jzitnik.chronos.repository.ItemRepository;
import cz.jzitnik.chronos.repository.PlayerRepository;
import cz.jzitnik.chronos.services.GameService;
import cz.jzitnik.chronos.utils.anotations.CheckUser;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

@CrossOrigin(origins = "*", maxAge = 3600)
@RestController
@RequestMapping("/game/players")
public class PlayerController {
    @Autowired
    private GameRepository gameRepository;

    @Autowired
    private PlayerRepository playerRepository;

    @Autowired
    private GameService gameService;

    @Autowired
    private ItemRepository itemRepository;

    @PostMapping("/new")
    public ResponseEntity<UnifiedResponse<String, Error>> newPlayer(@RequestParam String gameKey, @RequestBody PlayerNameRequest playerNameRequest) {
        var gameOptional = gameRepository.findByGameKey(gameKey);

        if (gameOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(
                UnifiedResponse.failure(new NotFoundError("Game with key " + gameKey + " was not found!"))
            );
        }

        var game = gameOptional.get();

        var player = gameService.addPlayer(game, playerNameRequest.getName());

        return ResponseEntity.ok(
                UnifiedResponse.success(player.getPlayerKey())
        );
    }

    @GetMapping("/me")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Player, Error>> getMe(@RequestParam String playerKey) {
        var player = playerRepository.findByPlayerKey(playerKey).get();

        return ResponseEntity.ok(UnifiedResponse.success(player));
    }

    @GetMapping("/me/inv_full")
    @CheckUser
    public ResponseEntity<UnifiedResponse<InventoryFullResponse, Error>> inventoryIsFull(@RequestParam String playerKey) {
        var player = playerRepository.findByPlayerKey(playerKey).get();

        return ResponseEntity.ok(UnifiedResponse.success(new InventoryFullResponse(player.getInventory().size() >= 3)));
    }

    @PostMapping("/use_item")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Object, Error>> useItem(@RequestParam String playerKey, @RequestParam Long itemId) {
        var player = playerRepository.findByPlayerKey(playerKey).get();
        var itemOptional = itemRepository.findById(itemId);

        if (itemOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Item was not found!")));
        }

        var item = itemOptional.get();

        if (!player.getInventory().contains(item)) {
            return ResponseEntity.status(HttpStatus.FORBIDDEN).body(UnifiedResponse.failure(new Error("You do not own this item!")));
        }

        // Use the item
        try {
            item.getItemType().useItem(player);
        } catch (ItemNotUsableException e) {
            return ResponseEntity.status(HttpStatus.BAD_REQUEST).body(UnifiedResponse.failure(e.toError()));
        }

        player.getMessages().add(new Message(player, String.valueOf(item.getId()), MessageType.ITEM_USED));

        playerRepository.save(player);

        item.setOwner(null);

        itemRepository.save(item);

        return ResponseEntity.ok(UnifiedResponse.success(null));
    }

    @GetMapping("/item")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Item, Error>> getItemInfo(@RequestParam String playerKey, @RequestParam Long itemId) {
        var itemOptional = itemRepository.findById(itemId);

        if (itemOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Item was not found!")));
        }
        var item = itemOptional.get();

        return ResponseEntity.ok(UnifiedResponse.success(item));
    }
}
